package parsers.CPA;

import java.util.*;
import util.*;
import core.*;
import parsers.CPA.*;

class CPACellProgram extends CellProgram
{
	private boolean resolved;
	Hashtable StateTable;

	CPACellProgram()
	{
		StateTable = new Hashtable();
		resolved = false;
	}

	void addState (CellState p_state)
	{
		StateTable.put(p_state.getState(),p_state);
	}

	void resolveStates ()
	{
	}

	public void runCellProgramForOneStep(Cell cell)
	{
		CellState cs = (CellState)StateTable.get(cell.getState());
		if(cs != null)
		{
			Enumeration enum = cs.getCellTransitionList().getList().elements();
			while(enum.hasMoreElements())
			{
				CellTransition ct = (CellTransition) enum.nextElement();
				ChemicalConditionList2 ccl = ct.getChemicalConditionList();

				Vector argVector = new Vector();
				argVector.add("Check Condition");
				argVector.add(ccl);
				Vector retVec = cell.provideService(argVector);

				if( ( (Boolean)retVec.elementAt(0) ).booleanValue() )
				{
					try
					{
						performTransition(cell,ct,(String)retVec.elementAt(1));
					}
					catch(ArrayIndexOutOfBoundsException e)
					{
						performTransition(cell,ct,null);
					}
					break;
				}
			}
		}
	}

	boolean isValidTransition(Cell cell, CellTransition ct)
	{
		ChemicalConditionList2 ccl = ct.getChemicalConditionList();
		Vector argVector = new Vector();
		argVector.add("Check Condition");
		argVector.add(ccl);
		Vector retVec = cell.provideService(argVector);

		return ( (Boolean)retVec.elementAt(0) ).booleanValue();
	}

	void performTransition(Cell cell, CellTransition ct, String givenDirection)
	{
		Vector argVector = new Vector();
		if(ct.isCellDeath)
		{
			argVector.add(new String("Cell Death"));
			argVector.add(givenDirection);
			argVector.add(ct.getCellEmissionList().getList());
			argVector.add(ct.getCellStorageList().getList());
			argVector.add(ct.getCellDiffusionList().getList());
			cell.provideService(argVector);
		}
		else if(ct.isCellDivision)
		{
			argVector.add(new String("Cell Division"));
			argVector.add(ct.getCellDivision().getState1());
			argVector.add(ct.getCellDivision().getState2());

			String direction = ct.getCellDivision().getDivisionType();

			if(direction.equals("same"))
				direction = givenDirection;
			else if(direction.equals("opposite"))
			{
				if(givenDirection == null)
					return;
				if(givenDirection.equals("X+")) direction = "X-";
				else if(givenDirection.equals("X-")) direction = "X+";
				else if(givenDirection.equals("Y+")) direction = "Y-";
				else if(givenDirection.equals("Y-")) direction = "Y+";
				else if(givenDirection.equals("Z+")) direction = "Z-";
				else if(givenDirection.equals("Z-")) direction = "Z+";
				else direction = " ";
			}

			argVector.add(direction);
			argVector.add(givenDirection);
			argVector.add(ct.getCellEmissionList().getList());
			argVector.add(ct.getCellStorageList().getList());
			argVector.add(ct.getCellDiffusionList().getList());
			cell.provideService(argVector);
		}
		else
		{
			argVector.add(new String("State Change"));
			argVector.add(ct.getCellStateChange().getState());
			argVector.add(givenDirection);
			argVector.add(ct.getCellEmissionList().getList());
			argVector.add(ct.getCellStorageList().getList());
			argVector.add(ct.getCellDiffusionList().getList());
			cell.provideService(argVector);
		}
	}

	public Vector provideService(Vector args)
	{
		if(args.elementAt(0).equals("Get State Color"))
		{
			Vector argVector = new Vector();
			CellState cs = (CellState)StateTable.get((String)args.elementAt(1));
			if(cs == null)
			{
				System.err.println("Undefined state: "+((String)args.elementAt(1)).toString());
				System.exit(1);
			}
			Color c = cs.getColor();
			if(c == null)
			{
				c = new Color(0.5,0.5,0.5);
			}
			argVector.add(c);
			return argVector;
		}
		else if(args.elementAt(0).equals("State Emissions"))
		{
			Vector argVector = new Vector();
			CellState cs = (CellState)StateTable.get((String)args.elementAt(1));
			if(cs == null)
			{
				System.err.println("Undefined state: "+((String)args.elementAt(1)).toString());
				System.exit(1);
			}
			Vector cel = cs.getCellEmissionList();
			Vector cdl = cs.getCellDiffusionList();
			argVector.add(cel);
			argVector.add(cdl);
			return argVector;
		}

		return null;
	}

	public void init(Vector args)
	{
	}

	public String toString()
	{
		StringBuffer retVal = new StringBuffer();

		Enumeration enum = StateTable.elements();
		while(enum.hasMoreElements())
		{
			retVal.append( ( (CellState)enum.nextElement() ).toString() );
		}
		return retVal.toString();
	}
}
