package sysmodels.simplebiology;

import java.io.*;
import java.util.*;
import javax.media.j3d.*;
import javax.vecmath.*;
import com.sun.j3d.utils.geometry.*;

import core.*;
import util.*;

class SimpleBiology extends SystemModel
{
	CellProgram cellProgram;
	CellDetails cellDetails[][][];
	Vector IDMappings;

	int currCellID;
	Hashtable statistics;
	int numCellsKilled;
	int numCellsDied;
	
	int offsetX, offsetY, offsetZ;
	int dx[], dy[], dz[];

	public SimpleBiology(Simulation p_sim)
	{
		super(p_sim);
		currCellID = 0;
		numCellsKilled = 0;
		numCellsDied = 0;
		
		offsetX = offsetY = offsetZ = 25;

		dx = new int[6];
		dy = new int[6];
		dz = new int[6];
		for(int i=0; i<6; i++)
		{
			dx[i] = dy[i] = dz[i] = 0;
		}
		dx[0] = dy[2] = dz[4] = 1;
		dx[1] = dy[3] = dz[5] = -1;
	}

	public void init(Vector args)
	{
		String modelOptions = (String) args.elementAt(0);
		cellProgram = getSimulation().getCellProgram();

		statistics = new Hashtable();
		IDMappings = new Vector();
		
		cellDetails = new CellDetails[50][50][50];
		for(int i=0; i<50; i++)
			for(int j=0; j<50; j++)
				for(int k=0; k<50; k++)
					cellDetails[i][j][k] = null;
	}

	public void runCellProgramForOneStep()
	{
		diffuseAllCellEmissions();
		Enumeration enum = IDMappings.elements();
		while(enum.hasMoreElements())
		{
			CellDetails cd = (CellDetails)enum.nextElement();

			if(cd.isNew())
			{
				cd.setAsOld();
			}
			else
			{
				//diffuseAllCellEmissions();
				//getSensedStoredDiffusedChemicalsForCell(cd);
				Cell cell = cd.getCell();
				cell.runCellProgramForOneStep();
			}
		}
		updateStatistics();
	}

	private void getSensedStoredDiffusedChemicalsForCell(CellDetails cd)
	{
		int x = cd.getX(); int y = cd.getY(); int z = cd.getZ();
		int i = 0;
		Vector argVector;

		cd.disposeSensedChemicals();
		while(i<6)
		{
			CellDetails neighbor = cellDetails[offsetX+x+dx[i]][offsetY+y+dy[i]][offsetZ+z+dz[i]];
			if(neighbor != null)
			{
				argVector = new Vector();
				argVector.add(new String("Emit Chemicals"));
				neighbor.getCell().provideService(argVector);
			}
			i++;
		}

		argVector = new Vector();
		argVector.add(new String("Store Chemicals"));
		cd.getCell().provideService(argVector);
		argVector = new Vector();
		argVector.add(new String("Sense Chemicals"));
		argVector.add(cd.getSensor("X+"));
		argVector.add(cd.getSensor("X-"));
		argVector.add(cd.getSensor("Y+"));
		argVector.add(cd.getSensor("Y-"));
		argVector.add(cd.getSensor("Z+"));
		argVector.add(cd.getSensor("Z-"));
		argVector.add(cd.getSensedChemicals());
		cd.getCell().provideService(argVector);
		cd.disposeSensedChemicals();
	}

	private void diffuseAllCellEmissions()
	{
		/* Cells emit chemicals depending upon their state */
		Enumeration enum = IDMappings.elements();
		while(enum.hasMoreElements())
		{
			CellDetails cd = (CellDetails)enum.nextElement();
			cd.disposeSensedChemicals();
		}

		enum = IDMappings.elements();
		while(enum.hasMoreElements())
		{
			CellDetails cd = (CellDetails)enum.nextElement();
			Cell cell = cd.getCell();
			Vector argVector = new Vector();
			argVector.add(new String("Emit Chemicals"));
			cell.provideService(argVector);
			argVector = new Vector();
			argVector.add(new String("Store Chemicals"));
			cell.provideService(argVector);
		}

		/* Now distribute the chemicals diffused during the run to individual cells */
		enum = IDMappings.elements();
		while(enum.hasMoreElements())
		{
			Vector argVector = new Vector();
			CellDetails cd = (CellDetails)enum.nextElement();
			Cell cell = cd.getCell();
			argVector.add(new String("Sense Chemicals"));
			argVector.add(cd.getSensor("X+"));
			argVector.add(cd.getSensor("X-"));
			argVector.add(cd.getSensor("Y+"));
			argVector.add(cd.getSensor("Y-"));
			argVector.add(cd.getSensor("Z+"));
			argVector.add(cd.getSensor("Z-"));
			argVector.add(cd.getSensedChemicals());
			cell.provideService(argVector);
			//System.err.println(cd.toString());
			cd.disposeSensedChemicals();
		}
	}
	
	public Vector provideService(Vector args)
	{
		if(args == null)
			return null;
		String service = (String)args.elementAt(0);
		if(service.equals("Diffuse Chemicals"))
		{
			int x, y, z;
			int id = ((Integer)args.elementAt(1)).intValue();
			Vector emissions = (Vector)args.elementAt(2);
			Vector diffusions = (Vector)args.elementAt(3);
			CellDetails cd1 = getCellDetails(id);

			x = (cd1.getX());
			y = (cd1.getY());
			z = (cd1.getZ());
			diffuseChemicals(x,y,z,diffusions);

			try
			{
				if(cellDetails[x+1+offsetX][y+offsetY][z+offsetZ] != null)
					cellDetails[x+1+offsetX][y+offsetY][z+offsetZ].diffuseChemicals(cd1,emissions);
				if(cellDetails[x-1+offsetX][y+offsetY][z+offsetZ] != null)
					cellDetails[x-1+offsetX][y+offsetY][z+offsetZ].diffuseChemicals(cd1,emissions);
				if(cellDetails[x+offsetX][y+1+offsetY][z+offsetZ] != null)
					cellDetails[x+offsetX][y+1+offsetY][z+offsetZ].diffuseChemicals(cd1,emissions);
				if(cellDetails[x+offsetX][y-1+offsetY][z+offsetZ] != null)
					cellDetails[x+offsetX][y-1+offsetY][z+offsetZ].diffuseChemicals(cd1,emissions);
				if(cellDetails[x+offsetX][y+offsetY][z+1+offsetZ] != null)
					cellDetails[x+offsetX][y+offsetY][z+1+offsetZ].diffuseChemicals(cd1,emissions);
				if(cellDetails[x+offsetX][y+offsetY][z-1+offsetZ] != null)
					cellDetails[x+offsetX][y+offsetY][z-1+offsetZ].diffuseChemicals(cd1,emissions);
			}
			catch(ArrayIndexOutOfBoundsException e)
			{
			}
		}
		if(service.equals("Store Chemicals"))
		{
			int id = ((Integer)args.elementAt(1)).intValue();
			Vector emissions = (Vector)args.elementAt(2);
			CellDetails cd = getCellDetails(id);

			storeChemicals(cd,emissions);
		}
		else if(service.equals("Cell Division"))
		{
			int id = ((Integer)args.elementAt(1)).intValue();
			String state1 = (String)args.elementAt(2);
			String state2 = (String)args.elementAt(3);
			String divisionType = (String)args.elementAt(4);
			performCellDivision(id,state1,state2,divisionType);
		}
		else if(service.equals("Cell Death"))
		{
			int id = ((Integer)args.elementAt(1)).intValue();
			Enumeration enum = IDMappings.elements();
			while(enum.hasMoreElements())
			{
				CellDetails cd = (CellDetails)(enum.nextElement());
				if(cd.getCell().getID().intValue() == id)
				{
					numCellsDied++;
					cellDetails[cd.getX()+offsetX][cd.getY()+offsetY][cd.getZ()+offsetZ] = null;
					IDMappings.remove(cd);
					break;
				}
			}
		}
		else if(service.equals("Simulation Instruction"))
		{
			Vector instruction = (Vector)args.elementAt(1);
			runInstruction(instruction);
		}
		else if(service.equals("Get Statistics"))
		{
			Vector argVector = new Vector();
			argVector.add(statistics);
			return argVector;
		}
		return null;
	}

	public BranchGroup getDisplayGraph()
	{
		int maxX = 1, maxY = 1, maxZ = 1, minX = -1, minY = -1, minZ = -1, largest;
		int x, y, z;
		// Globally used colors
		Color3f black = new Color3f(0.0f, 0.0f, 0.0f);
		Color3f white = new Color3f(1.0f, 1.0f, 1.0f);

		BranchGroup objRoot = new BranchGroup();

		/* Calculate the box size for display */
		Enumeration enum = IDMappings.elements();
		while(enum.hasMoreElements())
		{
			CellDetails cd = (CellDetails)enum.nextElement();
			x = cd.getX();
			y = cd.getY();
			z = cd.getZ();
			if(x > maxX) maxX = x;
			if(y > maxY) maxY = y;
			if(z > maxZ) maxZ = z;
			if(x < minX) minX = x;
			if(y < minY) minY = y;
			if(z < minZ) minZ = z;
		}
		maxX = (maxX > -minX)? maxX : -minX;
		maxY = (maxY > -minY)? maxY : -minY;
		maxZ = (maxZ > -minZ)? maxZ : -minZ;
		largest = (maxX > maxY) ? maxX : maxY;
		largest = (largest > maxZ)? largest : maxZ;
		float boxSize = 1.0f/largest;

		enum = IDMappings.elements();
		int curr=0;
		while(enum.hasMoreElements())
		{
			CellDetails cd = (CellDetails)enum.nextElement();
			Cell cell = cd.getCell();

			Appearance appearance = new Appearance();
			// Set up the material properties
			Color3f objColor = cell.getColor().getColor3fObject();
			appearance.setMaterial(new Material(black, black, objColor, white, 80.0f));
			appearance.setLineAttributes(new LineAttributes(2.0f, LineAttributes.PATTERN_SOLID, true));

			TransformGroup objTrans = new TransformGroup();
			Transform3D t = new Transform3D();
			Position cellPos = new Position(cd.getX()*boxSize,cd.getY()*boxSize,cd.getZ()*boxSize);
			t.set(cellPos.getVector3dObject());
			objTrans.setTransform(t);

			Node obj = new Box(boxSize/2, boxSize/2, boxSize/2,
								Box.GENERATE_NORMALS | Box.GENERATE_TEXTURE_COORDS, appearance);
			objTrans.addChild(obj);
			objRoot.addChild(objTrans);

			curr++;
		}

		return objRoot;
	}

	private CellDetails getCellDetails(int id)
	{
		Enumeration enum = IDMappings.elements();
		while(enum.hasMoreElements())
		{
			CellDetails cd = (CellDetails)enum.nextElement();

			if(id == cd.getCell().getID().intValue())
				return cd;
		}
		return null;
	}

	private void diffuseChemicals(double x1, double y1, double z1, Vector emissions)
	{
		double x2, y2, z2, dist;

		Enumeration enum = IDMappings.elements();
		while(enum.hasMoreElements())
		{
			CellDetails cd = (CellDetails)(enum.nextElement());
			x2 = cd.getX();
			y2 = cd.getY();
			z2 = cd.getZ();
			dist = Math.sqrt( (x1-x2)*(x1-x2) + (y1-y2)*(y1-y2) + (z1-z2)*(z1-z2) );

			Enumeration emissionEnum = emissions.elements();
			while(emissionEnum.hasMoreElements())
			{
				CellEmission emission = (CellEmission)emissionEnum.nextElement();
				double conc = emission.getConcentration() - dist;
				if(conc >= 0)
				{
					String chem = new String(emission.getChemical());
					CellEmission ce = new CellEmission(chem,conc,"all");
					cd.addSensedChemical(ce);
				}
			}
		}
	}

	private void storeChemicals(CellDetails cd, Vector emissions)
	{
		Enumeration emissionEnum = emissions.elements();
		while(emissionEnum.hasMoreElements())
		{
			CellEmission emission = (CellEmission)emissionEnum.nextElement();
			cd.addSensedChemical(emission);
		}
	}

	private CellDetails getNewCell(String state)
	{
		CellDetails cd = null;
		Integer ID = new Integer(currCellID);

		try
		{
			Vector args = new Vector();
			args.add(new String(state));
			Cell cell = getSimulation().createCell(ID, this, cellProgram, args);
			cd = new CellDetails(cell);
			IDMappings.add(cd);
			currCellID++;
		}
		catch(Exception e)
		{
			System.err.println(e);
		}

		return cd;
	}

	private void performCellDivision(int id, String state1, String state2, String divisionType)
	{
		CellDetails cd1 = null, cd2;
		Enumeration enum = IDMappings.elements();
		while(enum.hasMoreElements())
		{
			CellDetails cd = (CellDetails)enum.nextElement();
			if(cd.getCell().getID().intValue() == id)
			{
				cd1 = cd;
				break;
			}
		}
		cd2 = getNewCell(state2);
		if(cd1 == null)
		{
			System.err.println("Non-existant cell requested cell division?");
			System.exit(1);
		}
		if(cd2 == null)
		{
			System.err.println("Not enough Memory");
			System.exit(2);
		}
		if(divisionType == null)
		{
			System.err.println("Unable to perform cell division");
			System.err.println("("+state1+", "+state2+")");
			return;
		}

		cd1.getCell().setState(state1);

		int x, y, z, dx = 0, dy = 0, dz = 0;
		x = cd1.getX();
		y = cd1.getY();
		z = cd1.getZ();
		if(divisionType.equals("X+")) dx = 1;
		else if(divisionType.equals("X-")) dx = -1;
		else if(divisionType.equals("Y+")) dy = 1;
		else if(divisionType.equals("Y-")) dy = -1;
		else if(divisionType.equals("Z+")) dz = 1;
		else if(divisionType.equals("Z-")) dz = -1;

		cd2.setPosition(x+dx,y+dy,z+dz);
		displaceCells(cd2,dx,dy,dz);

		cd2.setAsNew();
	}

	private void displaceCells (CellDetails cd, int dx, int dy, int dz)
	{
		boolean displaced = true;
		CellDetails cdOther;
		int x, y, z;

		while(displaced)
		{
			x = cd.getX();
			y = cd.getY();
			z = cd.getZ();
		
			cdOther = cellDetails[x+offsetX][y+offsetY][z+offsetZ];
			if(cdOther == null)
			{
				displaced = false;
				cellDetails[x+offsetX][y+offsetY][z+offsetZ] = cd;
			}
			else
			{
				displaced = true;
				cdOther.setPosition(x+dx,y+dy,z+dz);
				cellDetails[x+offsetX][y+offsetY][z+offsetZ] = cd;
				cd = cdOther;
			}
		}
	}

	private void runInstruction(Vector instruction)
	{
		String instrType = (String)instruction.elementAt(0);
		//System.err.println(instrType);
		if(instrType.equals("new"))
		{
			double x = ((Double)instruction.elementAt(2)).doubleValue();
			double y = ((Double)instruction.elementAt(3)).doubleValue();
			double z = ((Double)instruction.elementAt(4)).doubleValue();
			CellDetails cd = getNewCell((String)instruction.elementAt(1));
			cd.setPosition(x, y, z);
			cellDetails[(int)x+offsetX][(int)y+offsetY][(int)z+offsetZ] = cd;
		}
		else if(instrType.equals("kill"))
		{
			double x1, y1, z1, x2, y2, z2, radius;
			String area = (String)instruction.elementAt(1);
			if(area.equals("sphere"))
			{
				x1 = ((Double)instruction.elementAt(2)).doubleValue();
				y1 = ((Double)instruction.elementAt(3)).doubleValue();
				z1 = ((Double)instruction.elementAt(4)).doubleValue();
				radius = ((Double)instruction.elementAt(5)).doubleValue();
				killSphere(x1, y1, z1, radius);
			}
			if(area.equals("block"))
			{
				x1 = ((Double)instruction.elementAt(2)).doubleValue();
				y1 = ((Double)instruction.elementAt(3)).doubleValue();
				z1 = ((Double)instruction.elementAt(4)).doubleValue();
				x2 = ((Double)instruction.elementAt(5)).doubleValue();
				y2 = ((Double)instruction.elementAt(6)).doubleValue();
				z2 = ((Double)instruction.elementAt(7)).doubleValue();
				killBlock(x1, y1, z1, x2, y2, z2);
			}
		}
		else if(instrType.equals("Output"))
		{
			String s = null, file = null;
			
			try
			{
				file = (String)instruction.elementAt(1);
			}
			catch(ArrayIndexOutOfBoundsException e)
			{
				System.err.println(e.toString());
			}
			System.out.println("Here: " + file);
			BufferedWriter outputWriter = null;

			try
			{
				if(file != null)
					outputWriter = new BufferedWriter(new FileWriter(file));

				Enumeration enum = IDMappings.elements();
				while(enum.hasMoreElements())
				{
					CellDetails cd = (CellDetails)enum.nextElement();
					s = cd.getX() + " " + cd.getY() + " " + cd.getZ() + " " +
						cd.getCell().getColor().getR() + " " +
						cd.getCell().getColor().getG() + " " +
						cd.getCell().getColor().getB() + " ";
					if(outputWriter == null)
						getSimulation().writeOutputLine(s);
					else
					{
						outputWriter.write(s,0,s.length());
						outputWriter.newLine();
					}
				}
				if(outputWriter != null)
					outputWriter.close();
			}
			catch(IOException e)
			{
				System.err.println(e.toString());
			}
		}
		else if(instrType.equals("Statistics"))
		{
			try
			{
				Enumeration enum = IDMappings.elements();
				while(enum.hasMoreElements())
				{
					CellDetails cd = (CellDetails)enum.nextElement();
					getSimulation().writeOutputLine(cd.getX() + " " +
													cd.getY() + " " +
													cd.getZ() + " " +
													cd.getCell().getColor().getR() + " ");
				}
			}
			catch(IOException e) {}
		}
	}

	private void killSphere(double x1, double y1, double z1, double radius)
	{
		double x2, y2, z2;
		boolean killed = true;
		while(killed)
		{
			killed = false;
			Enumeration enum = IDMappings.elements();
			while(enum.hasMoreElements())
			{
				double dist;
				CellDetails cd = (CellDetails)enum.nextElement();
				x2 = cd.getX();
				y2 = cd.getY();
				z2 = cd.getZ();
				dist = Math.sqrt( (x1-x2)*(x1-x2) + (y1-y2)*(y1-y2) + (z1-z2)*(z1-z2) );
				if(dist <= radius)
				{
					numCellsKilled++;
					killed = true;
					IDMappings.remove(cd);
					cellDetails[(int)x2+offsetX][(int)y2+offsetY][(int)z2+offsetZ] = null;
				}
			}
		}
	}

	private void killBlock(double x1, double y1, double z1, double x2, double y2, double z2)
	{
		double x, y, z;
		boolean killed = true;
		while(killed)
		{
			killed = false;
			Enumeration enum = IDMappings.elements();
			while(enum.hasMoreElements())
			{
				CellDetails cd = (CellDetails)enum.nextElement();
				x = cd.getX(); y = cd.getY(); z = cd.getZ();
				if( (x1 <= x) && (x <= x2) && (y1 <= y) && (y <= y2) && (z1 <= z) && (z <= z2) ) 
				{
					numCellsKilled++;
					killed = true;
					IDMappings.remove(cd);
					cellDetails[(int)x+offsetX][(int)y+offsetY][(int)z+offsetZ] = null;
				}
			}
		}
	}

	private void updateStatistics()
	{
		statistics.put(new String("Simulation Step"), (new Integer(getSimulation().getStep())).toString());
		statistics.put(new String("Number of cells"), (new Integer(IDMappings.size())).toString());
		statistics.put(new String("Number of cells killed"), (new Integer(numCellsKilled)).toString());
		statistics.put(new String("Number of cells died"), (new Integer(numCellsDied)).toString());
	}

	static class SimpleBiologyFactory extends SystemFactory
	{
		public SystemModel create(Simulation p_sim, String modelOptions)
		{
			SimpleBiology sb = new SimpleBiology(p_sim);

			Vector vec = new Vector();
			vec.add(modelOptions);
			sb.init(vec);

			return sb;
		}
	}

	static
	{
		Simulation.sysModels.put(new String("sysmodels.simplebiology.SimpleBiology"),
								 new SimpleBiologyFactory());
	}
}
